package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.citygml4j.core.model.core.AbstractSpace;
import org.citygml4j.core.model.transportation.AbstractTransportationSpace;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;

import java.util.ArrayList;
import java.util.List;

public abstract class TransportationSpace extends TransportationObject {

    private final List<TrafficSpaceObject> trafficSpaces = new ArrayList<>();
    private final List<TrafficSpaceObject> auxTrafficSpaces = new ArrayList<>();

    protected TransportationSpace(TransportationType type) {
        super(type);
    }

    public List<TrafficSpaceObject> getTrafficSpaces() {
        return trafficSpaces;
    }

    public List<TrafficSpaceObject> getAuxTrafficSpaces() {
        return auxTrafficSpaces;
    }

    public void addTrafficSpace(TrafficSpaceObject t) {
        trafficSpaces.add(t);
    }

    public void addAuxTrafficSpace(TrafficSpaceObject t) {
        auxTrafficSpaces.add(t);
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        for (Geometry geom : getGeometries()) {
            if (geom instanceof ImplicitGeometryHolder) {
                continue;
            }
            if (geom.getType() == GeometryType.MULTI_SURFACE) {
                MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
                if (super.getGmlObject() instanceof AbstractTransportationSpace ats) {
                    setMultiSurfaceAccordingToLod(ats, ms, geom.getLod());
                } else {
                    AbstractSpace as = (AbstractSpace) super.getGmlObject();
                    setMultiSurfaceAccordingToLod(as, ms, geom.getLod());
                }
            } else {
                throw new IllegalStateException("Geometry in TransportationObject cannot be of type " + geom.getType()
                        + ". Only MultiSurface allowed");
            }
        }
        for (TrafficSpaceObject spaces : trafficSpaces) {
            spaces.reCreateGeometries(factory, config);
        }
        for (TrafficSpaceObject spaces : auxTrafficSpaces) {
            spaces.reCreateGeometries(factory, config);
        }
    }

    private void setMultiSurfaceAccordingToLod(AbstractTransportationSpace ats, MultiSurface ms, Lod lod) {
        switch (lod) {
            case LOD0 -> ats.setLod0MultiSurface(new MultiSurfaceProperty(ms));
            case LOD1 -> ats.getDeprecatedProperties().setLod1MultiSurface(new MultiSurfaceProperty(ms));
            case LOD2 -> ats.setLod2MultiSurface(new MultiSurfaceProperty(ms));
            case LOD3 -> ats.setLod3MultiSurface(new MultiSurfaceProperty(ms));
            case LOD4 -> ats.getDeprecatedProperties().setLod4MultiSurface(new MultiSurfaceProperty(ms));
            default ->
                    throw new IllegalStateException("cannot set geometry with LOD for AuxiliaryTrafficSpace: " + lod);
        }
    }

    private void setMultiSurfaceAccordingToLod(AbstractSpace as, MultiSurface ms, Lod lod) {
        switch (lod) {
            case LOD0:
                as.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                as.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                as.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException("cannot set geometry with LOD for AuxiliaryTrafficSpace: " + lod);
        }
    }


    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (TrafficSpaceObject to : trafficSpaces) {
            if (to.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            if (to.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (TrafficSpaceObject to : trafficSpaces) {
            to.clearAllContainedCheckResults();
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.clearAllContainedCheckResults();
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (TrafficSpaceObject to : trafficSpaces) {
            to.collectContainedErrors(errors);
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.collectContainedErrors(errors);
        }
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (TrafficSpaceObject to : trafficSpaces) {
            if (to.containsAnyError()) {
                return true;
            }
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            if (to.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (TrafficSpaceObject to : trafficSpaces) {
            to.accept(c);
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.accept(c);
        }
    }


    @Override
    public void unsetGmlGeometries() {
        AbstractSpace ats = (AbstractSpace) super.getGmlObject();
        ats.setLod0MultiSurface(null);
        ats.setLod2MultiSurface(null);
        ats.setLod3MultiSurface(null);
        for (TrafficSpaceObject to : trafficSpaces) {
            to.unsetGmlGeometries();
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.unsetGmlGeometries();
        }
    }

    @Override
    public CityObject getTopLevelCityObject() {
        return this;
    }


    @Override
    public String toString() {
        return "TransportationSpace [id=" + getGmlId() + "]";
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (TrafficSpaceObject to : trafficSpaces) {
            to.prepareForChecking();
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (TrafficSpaceObject to : trafficSpaces) {
            to.clearMetaInformation();
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.clearMetaInformation();
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        for (TrafficSpaceObject to : trafficSpaces) {
            to.collectInstances(handler);
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.collectInstances(handler);
        }
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        TransportationObject originalTo = (TransportationObject) original;
        for (TrafficSpaceObject to : trafficSpaces) {
            to.fillValues(originalTo, handler);
        }
        for (TrafficSpaceObject to : auxTrafficSpaces) {
            to.fillValues(originalTo, handler);
        }
    }


}
