/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 *
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check;

import de.hft.stuttgart.citydoctor2.CityDoctorValidation;
import de.hft.stuttgart.citydoctor2.checkresult.CheckReport;
import de.hft.stuttgart.citydoctor2.checkresult.GlobalErrorStatistics;
import de.hft.stuttgart.citydoctor2.checkresult.GlobalStatistics;
import de.hft.stuttgart.citydoctor2.checkresult.ModelStatistics;
import de.hft.stuttgart.citydoctor2.checkresult.utility.CheckReportParseException;
import de.hft.stuttgart.citydoctor2.checks.util.FeatureCheckedListener;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.exceptions.CityDoctorWriteException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipArchive;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipEntry;
import org.apache.commons.io.FileUtils;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

/**
 * @author Matthias Betz
 */
public class CheckerTest {

    @Rule
    public TemporaryFolder folder = new TemporaryFolder();

    @Test
    public void testSchematron() throws CityGmlParseException, InvalidGmlFileException {
        ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
        config.getRequirements().get(Requirement.R_SE_BS_ROOF_UNFRAGMENTED.toString()).setEnabled(false);
        config.setSchematronFilePathInGlobalParameters("src/test/resources/schematronTest.xml");
        CityDoctorModel model = CityGmlParser.parseCityGmlFile(
                "src/test/resources/SimpleSolid_SrefBS_SchematronTest.gml", config.getParserConfiguration());
        Checker checker = new Checker(config, model);
        checker.runChecks();
        for (Building b : model.getBuildings()) {
            if (b.getGmlId().getGmlString().equals("_Simple_BD.1")) {
                assertTrue(b.containsAnyError());
            } else {
                assertFalse(b.containsAnyError());
            }
        }
        assertFalse(model.getGlobalErrors().isEmpty());
    }

    @Test
    public void testChecker() throws CityGmlParseException, IOException, InvalidGmlFileException, CityDoctorWriteException {

        File f = folder.newFile();
        File f2 = folder.newFile();
        try {
            String[] args = new String[6];
            args[0] = "-in";
            args[1] = "src/test/resources/QA-CS-CONCOMP.gml";
            args[2] = "-xmlReport";
            args[3] = f.getAbsolutePath();
            args[4] = "-pdfReport";
            args[5] = f2.getAbsolutePath();
            CityDoctorValidation.main(args);
            assertTrue(f.exists());
            assertTrue(f2.exists());
        } finally {
            f.delete();
            f2.delete();
        }
    }

    @Test
    public void testStreaming() throws CityGmlParseException, IOException, InvalidGmlFileException, CityDoctorWriteException {
        File f = folder.newFile();
        File f2 = folder.newFile();
        File f3 = folder.newFile();
        try {
            String[] args = new String[10];
            args[0] = "-in";
            args[1] = "src/test/resources/testarea.gml";
            args[2] = "-config";
            args[3] = "src/test/resources/testConfigWithStreaming.yml";
            args[4] = "-pdfReport";
            args[5] = f.getAbsolutePath();
            args[6] = "-xmlReport";
            args[7] = f2.getAbsolutePath();
            args[8] = "-out";
            args[9] = f3.getAbsolutePath();
            CityDoctorValidation.main(args);
            assertTrue(f.exists());
            assertTrue(f2.exists());
            assertTrue(f3.exists());
        } finally {
            f.delete();
            f2.delete();
            f3.delete();
        }
    }

    @Test
    public void testCliZipChecking() throws IOException, CityDoctorWriteException, CityGmlParseException, InvalidGmlFileException {
        String output = folder.getRoot().toPath() + File.separator + "test";
        String pdfOutput = folder.getRoot().toPath() + File.separator + "pdf" + File.separator;
        String xmlOutput = folder.getRoot().toPath() + File.separator + "xml" + File.separator;
        try {
            String[] args = new String[10];
            args[0] = "-in";
            args[1] = "src/test/resources/zipArchive.zip";
            args[2] = "-config";
            args[3] = "src/test/resources/testConfigWithStreaming.yml";
            args[4] = "-pdfReport";
            args[5] = pdfOutput;
            args[6] = "-xmlReport";
            args[7] = xmlOutput;
            args[8] = "-out";
            args[9] = output + File.separator;
            CityDoctorValidation.main(args);
            File zipFile = new File(output + ".zip");
            File pdfDir = new File(pdfOutput);
            File xmlDir = new File(xmlOutput);
            assertTrue(zipFile.exists());
            assertTrue(pdfDir.exists());
            assertTrue(xmlDir.exists());

            assertTrue(pdfDir.isDirectory());
            assertTrue(xmlDir.isDirectory());
            assertEquals(5, pdfDir.listFiles().length);
            assertEquals(5, xmlDir.listFiles().length);

            CityGmlZipArchive cgmlArch = CityGmlZipArchive.register(output + ".zip");
            assertNotNull(cgmlArch);
            cgmlArch.mountArchive(new ParserConfiguration(8, true));
            assertEquals(5, cgmlArch.getEntries().size());
            for (CityGmlZipEntry entry : cgmlArch.getEntries()) {
                assertNotNull(entry);
                assertNull(entry.getErrorType());
            }

        } catch (CityGmlParseException | IOException | InvalidGmlFileException | CityDoctorWriteException e) {
            throw e;
        }
    }

    @Test
    public void testZipEntryChecking() throws CityGmlParseException, IOException {
        CityGmlZipArchive cgmlArch = CityGmlZipArchive.register("src/test/resources/zipArchive.zip");
        assertNotNull(cgmlArch);
        cgmlArch.mountArchive(new ParserConfiguration(8, false));
        ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
        FeatureCheckedListener l = co -> {
            assertTrue(co.isValidated());
        };
        for (CityGmlZipEntry entry : cgmlArch.getEntries()) {
            Checker.streamCheck(entry, null, null, config, l, null);
        }
    }

    @Test
    public void testZipArchiveChecking() throws CityGmlParseException, IOException {
        CityGmlZipArchive cgmlArch = CityGmlZipArchive.register("src/test/resources/zipArchive.zip");
        assertNotNull(cgmlArch);
        cgmlArch.mountArchive(new ParserConfiguration(8, false));
        ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
        FeatureCheckedListener l = co -> {
            assertTrue(co.isValidated());
        };
        Checker.streamCheck(cgmlArch, null, null, config, l, null);
    }

    @Test
    public void testReportCreation() throws IOException, CityGmlParseException, InvalidGmlFileException {
        Path tmpDir = null;
        try {
            tmpDir = Files.createTempDirectory("testTmp");
            tmpDir.toFile().deleteOnExit();
            Path xmlPath = tmpDir.resolve("report.xml");
            Path pdfPath = tmpDir.resolve("report.pdf");
            File features = new File("src/test/resources/feature_types.gml");
            CityDoctorModel model = CityGmlParser.parseCityGmlFileSilently(
                    features.toString(), new ParserConfiguration(8, false));
            assertNotNull(model);
            Checker checker = new Checker(model);
            checker.runChecks();
            checker.writeXmlReport(xmlPath.toString());
            checker.writePdfReport(pdfPath.toString());
            assertTrue(Files.exists(xmlPath));
            assertTrue(Files.exists(pdfPath));
            try {
                CheckReport report = CheckReport.load(xmlPath.toString());
                assertNotNull(report);

                GlobalStatistics globalStatistics = report.getGlobalStatistics();
                assertNotNull(globalStatistics);

                ModelStatistics modelStats = globalStatistics.getModelStats();
                assertNotNull(modelStats);
                assertEquals(1, modelStats.getNumBuildings());
                assertEquals(1, modelStats.getNumVegetation());
                assertEquals(1, modelStats.getNumTransportation());
                assertEquals(1, modelStats.getNumTunnelObjects());
                assertEquals(1, modelStats.getNumBridgeObjects());
                assertEquals(1, modelStats.getNumWaterObjects());
                assertEquals(1, modelStats.getNumLandObjects());
                assertEquals(1, modelStats.getNumCityFurniture());
                assertEquals(1, modelStats.getNumGenericCityObjects());

                GlobalErrorStatistics globError = globalStatistics.getGlobalErrorStats();
                assertNotNull(globError);
                assertEquals(0, globError.getNumErrorBuildings());
                assertEquals(0, globError.getNumErrorVegetation());
                assertEquals(0, globError.getNumErrorTransportation());
                assertEquals(0, globError.getNumErrorTunnelObjects());
                assertEquals(0, globError.getNumErrorBridgeObjects());
                assertEquals(0, globError.getNumErrorWaterObjects());
                assertEquals(0, globError.getNumErrorLandObjects());
                assertEquals(1, globError.getNumErrorCityFurniture());
                assertEquals(1, globError.getNumErrorGenericCityObjects());


            } catch (CheckReportParseException e) {
                fail("Report should be valid file");
            }


        } finally {
            if (tmpDir != null) {
                FileUtils.deleteDirectory(tmpDir.toFile());
            }
        }
    }

}
